class Collector {
    constructor(){
        this.setup()
    }
    /**
     * Esto es fundamental para el codigo, si lo usas algunos datos pueden resetearse
     * @deprecated
     */
    setup(){
        /**
         * Los keys del Collector
         * @return Collector.Keys:Array
         */
        this.keys = new Array()
        /**
         * Los values del Collector
         * @return Collector.Values:Array
         */
        this.values = new Array()
        /**
         * La lista del codigo (Para ti no sirve comparado con las otras funciones)
         * @deprecated
         */
        this.list = new Array()
    }
    /**
     * Estableces un dato en el Collector
     * @param {String} _key 
     * @param {*} _value 
     * @returns Collector Functions:Functions
     * @example Collector.set("Key", ['Value one', 'Value two'])
     */
    set(_key,_value){
        if(!_key) 
            throw new Error(`Incluya el key`)
        if(typeof _key !== "string") 
            throw new Error(`La key solo puede ser en string`)
        if(!_value) 
            throw new Error(`Incluya el value del key "${_key}"`)
        if(this.keys.indexOf(_key) > 0)
            throw new Error(`Este dato ya existe en la collection`)
        let _json;
        _json = {key:_key,value:_value}
        this.list.push(_json)
        this.keys.push(_key)
        this.values.push(_value)
        return this
    }
    /**
     * Eliminas un dato del Collector
     * @param {String} key 
     * @returns Collector Functions:Functions
     * @example Collector.remove("Key")
     */
    remove(...key){
        if(!key)
            throw new Error("Incluya el key")
        let keys = new Array()
        for(let _key of key){
            keys.push(_key)
        }
        for(let _key in keys){
            let index = this.keys.indexOf(keys[_key])
            if(index < 0)
                throw new Error(`Dato no encontrado`)
            this.list.splice(index,1)
            this.keys.splice(index,1)
            this.values.splice(index,1)
        }
        return this
    }
    /**
     * Reseteas el Collector
     * @returns Collector Functions:Functions
     * @example Collector.reset()
     */
    reset(){
        this.list = new Array()
        this.keys = new Array()
        this.values = new Array()
        return this
    }
    /**
     * Mapeas los datos del Collector
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.Map:Array
     * @example Collector.map(x => `Key: ${x.key} | Value: ${x.value}`)
     */
    map(fn,args){
        if(typeof args !== 'undefined')
            fn = fn.bind(args)
        let _map = new Array()
        let $map = new Array()
        for(var key in this.list){
            _map.push(fn(this.list[key]))
            $map = _map
        }
        _map = new Array()
        return $map
    }
    /**
     * Obtienes el valor de un key que este dentro del Collector
     * @param {String} key 
     * @returns Key.Values:TypeOfValue
     * @example Collector.get("Key")
     */
    get(key){
        if(!key)
            throw new Error("Incluya el key")
        let index = this.keys.indexOf(key)
        if(index < 0)
            throw new Error(`Dato no encontrado`)
        return this.values[index]
    }
    /**
     * Mapeas los valores del Collector
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.MapValues:Array
     * @example Collector.mapValues(x => `Value: ${x}`)
     */
    mapValues(fn,args){
        if(typeof args !== "undefined")
            fn = fn.bind(args)
        let _mapValues = new Array()
        let $mapValues = new Array()
        for(let key in this.list){
            let $json = this.list[key]
            _mapValues.push({key: $json.key, value: fn($json.value)})
            $mapValues = _mapValues
        }
        _mapValues = new Array()
        return $mapValues
    }
    /**
     * Mapeas los key del Collector
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.MapKeys:Array
     * @example Collector.mapKeys(x => `Key: ${x}`)
     */
    mapKeys(fn,args){
        if(typeof args !== "undefined")
            fn = fn.bind(args)
        let _mapKeys = new Array()
        let $mapKeys = new Array()
        for(let key in this.list){
            let $json = this.list[key]
            _mapKeys.push({key: fn($json.key), value: $json.value})
            $mapKeys = _mapKeys
        }
        _mapKeys = new Array()
        return $mapKeys
    }
    /**
     * Buscas un key que cumpla con algo
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.FindKey:TypeOfValue
     * @example Collector.findKey(x => x.includes("K"))
     */
    findKey(fn,args){
        if (typeof args !== 'undefined')
            fn = fn.bind(args)
        for (let key in this.list) {
            if(fn(this.list[key].key))
                return this.list[key]
        }
        return undefined
    }
    /**
     * Buscas un value que cumpla con algo
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.FindValue:TypeOfValue
     * @example Collector.findValue(x => x.id === 12)
     */
    findValue(fn,args){
        if (typeof args !== 'undefined')
            fn = fn.bind(args)
        for (let key in this.list) {
            if(fn(this.list[key].value))
                return this.list[key]
        }
        return undefined
    }
    /**
     * Revisas si todos los datos del Collector cumplen con algo
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.Every:Boolean
     * @example Collector.every(x => x.key === "Key")
     */
    every(fn,args){
        if(typeof args !== "undefined")
            fn = fn.bind(args)
        for(let pkey in this.list){
            if (!fn(this.list[pkey]))
                return false
        }
        return true 
    }
    /**
     * Revisas si algunos de los datos del Collector cumplen con algo
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.Some:Boolean
     * @example Collector.some(x => x.key === "Key")
     */
    some(fn,args){
        if(typeof args !== "undefined")
            fn = fn.bind(args)
        for(let pkey in this.list){
            if (fn(this.list[pkey]))
                return true
        }
        return false
    }
    /**
     * Clonas este Collector
     * @returns Collector.Constructor:Constructor
     * @example Collector.clone()
     */
    clone(){
        let Collector = new this.constructor()
        let clone = this.list
        Collector.setup()
        for(let key in clone){
            let _key = clone[key].key
            let _value = clone[key].value
            Collector.set(_key,_value)
        }
        return Collector
    }
    /**
     * El primer dato del Collector
     * @returns Collector.Keys && Collector.Values:Object
     * @example Collector.first()
     */
    first(){
        return this.list[0]
    }
    /**
     * El ultimo dato del Collector
     * @returns Collector.Keys && Collector.Values:Object
     * @example Collector.last()
     */
    last(){
        return this.list[this.list.length-1]
    }
    /**
     * Buscas algun dato del Collector
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.Find:Object
     * @example Collector.find(x => x.key === "Key")
     */
    find(fn,args){
        if(typeof args !== "undefined")
            fn = fn.bind(args)
        for(let key in this.list){
            if(fn(this.list[key]))
                return this.list[key]
        }
        return undefined
    }
    /**
     * Buscas algun dato del Collector y lo eliminas
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.FindAndDelete:Constructor
     * @example Collector.findAndDelete(x => x.key.includes("K"))
     */
    findAndDelete(fn,args){
       let _delete = this.find(fn,args)
       let $delete = _delete === undefined ? true : false
       if($delete)
            return undefined
        return this.remove(_delete.key)
    }
    /**
     * Te retorna los datos que cumplen el filtro que quieras
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.Filter:Array
     * @example Collector.filter(x => x.key.includes("K"))
     */
    filter(fn,args){
        if (typeof args !== 'undefined')
            fn = fn.bind(args);
        let Constructor = new this.constructor()
        Constructor.setup()
        for(let key in this.list){
            if(fn(this.list[key]))
                Constructor.set(this.list[key].key,this.list[key].value)
        }
        return Constructor
    }
    /**
     * Te da un dato random del Collector
     * @returns Collector.Random:Object
     * @example Collector.random()
     */
    random(){
        return this.list[Math.floor(Math.random()*this.list.length)]
    }
    /**
     * Te da un key random del Collector
     * @returns Collector.RandomKey:String
     * @example Collector.randomKey()
     */
    randomKey(){
        return this.keys[Math.floor(Math.random()*this.keys.length)]
    }
    /**
     * Te da un value random del Collector
     * @returns Collector.RandomValue:TypeOfValue
     * @example Collector.randomValue()
     */
    randomValue(){
        return this.values[Math.floor(Math.random()*this.values.length)]
    }
    /**
     * Te dice si el dato existe en el Collector
     * @returns Collector.Has:Boolean
     * @example Collector.has("Key")
     */
    has(key){
        if(!key)
            throw new Error("Incluya el key")
        return this.keys.includes(key)
    }
    /**
     * Ordenas los datos del Collector (solo funciona con numeros)
     * @param {*} options
     * @returns Collector.Sort:Collector
     * @example Collector.sort((nm,nm2) => nm - nm2)
     */
    sort(options){
        let $sort = new Array()
        for(var key in this.list){
            $sort.push(this.list[key].value)
        }
        let _sort = $sort.sort(options)
        let Collector = new this.constructor()
        Collector.setup()
        for(var key in _sort){
            let _value = _sort[key]
            let _key = this.find(x => x.value === _value)
            Collector.set(_key.key,_key.value)
        }
        this.list = Collector.list
        this.keys = Collector.keys 
        this.values = Collector.values
        return Collector
    }/**
     * Obtienes todos los datos del Collector
     * @param {*} options
     * @returns Collector.ForEach:Collector
     * @example Collector.forEach((x) => console.log(x))
     */
    forEach(fn,args){
        if(typeof args !== "undefined")
            fn = fn.bind(args)
        let Collector = this.clone()
        return fn(Collector)
    }
    /**
     * Revisas si un Collector es el mismo que este collector (si tienen los mismos datos)
     * @param {Collection} Collection
     * @returns Collector.Equals:Boolean
     * @example Collector.equals(new Collection().set("key","10")) 
     */
    equals(Collection){
        if(!Collection)
            return false
        if(this.list === Collection.list)
            return true
        if(this.list.length !== Collection.list.length) 
            return false
        for(let key in this.list){
            if (!Collection.has(this.list[key].key)){
                return false
            } else {
                if(typeof this.list[key].value === "object"){
                    for(var value in Collection.values[key]){
                        for(var $value in this.values[key]){
                            if(value !== $value)
                                return false
                            if(this.values[key][$value] !== Collection.values[key][value])
                                return false
                        }
                    }
                } else if(this.list[key].value !== Collection.get(this.list[key].key)) {
                    return false
                }
            }
        }
        return true
    }
    /**
     * Actualizas el value de algun key
     * @param {String} _key
     * @param {*} _newValue
     * @returns Collector.Update:Collector
     * @example Collector.update("Key",10)
     */
    update(_key,_newValue){
        if(!_key)
            throw new Error(`Inculya el key`)
        if(!_newValue)
            throw new Error(`Incluya el nuevo value del key`)
        let has = this.has(_key)
        if(!has)
            throw new Error(`Key no encontrada`)
        let Collector = new this.constructor()
        let update = this.list
        Collector.setup()
        for(let key in update){
            let $key = update[key].key
            let $value = update[key].value
            if($key === _key) $value = _newValue
            Collector.set($key,$value)
        }
        this.list = Collector.list
        this.keys = Collector.keys
        this.values = Collector.values
        return Collector
    }
    /**
     * Combinas Collectors para obtener uno solo
     * @param {Function} fn 
     * @param {Function|Arguments} args 
     * @returns Collector.Concat:Constructor
     * @example Collector.concat(new Collection().set('key',"09"))
     */
    concat(...Collection){
        let Collector = this.clone()
        for(let collector of Collection){
            for(let key in collector.list){
                Collector.set(collector.list[key].key,collector.list[key].value)
            }
        }
        return Collector
    }
    /**
     * Te da la posicion de un elemento (el index)
     * @param {String} key 
     * @returns Collector.Position:Number
     * @example Collector.position("Key")
     */
    position(key){
        if(!key)
            throw new Error(`Incluya la key`)
        return this.keys.indexOf(key)
    }
    /**
     * Revisas si el Collector tiene todas las keys que pones
     * @param  {String} key 
     * @returns Collector.HasAll:Boolean
     * @example Collector.hasAll("Test","Test2") //retorna false ya que estos dato no existen en el Collector
     */
    hasAll(...key){
        let _datos = new Array()
        for(let _key of key){
            _datos.push(_key)
        }
        if(_datos !== this.keys)
            return false
        return true
    }
    /**
     * Revisas si el Collector tiene algunos de las keys que pones
     * @param  {String} key 
     * @returns Collector.HasAny:Boolean
     * @example Collector.hasAny("Key","Test") //retorna true ya que si existe el key "Key"
     */
    hasAny(...key){
        let _boolean = new Array()
        let _false = new Array()
        let boolean;
        for(let _key of key){
            boolean = this.has(_key) ? true : false
            _false.push(false)
            _boolean.push(boolean)
        }
        return _boolean.join(' ') === _false.join(' ') ? false : true
    }
    /**
     * Revisas si una key tiene los mismos values que otra
     * @param {String} key 
     * @returns Collector.EqualsValues:Boolean
     * @example Collector.equalsValues("Key","Test")
     */
    equalsValues(key,key2){
        if(typeof key !== "string")
            throw new Error(`Key invalida`)
        if(typeof key2 !== "string")
            throw new Error(`Key invalida`)
        let _value = this.get(key)
        let $value = this.get(key2)
        if(typeof _value === "object" && typeof $value === "object"){
            for(var _key in _value){
                for(var __key in $value){
                    if(_key !== __key)
                        return false
                    if(_value[_key] !== $value[__key])
                        return false
                }
            }
        } else if(_value !== $value){
            return false
        }
        return true
    }
    /**
     * Te da todos los datos en un formato json o Object
     * @returns Collector.toJson:Object
     * @example Collector.toJson()
     */
    toJson(){
        return {
            all: this.list,
            keys: this.keys,
            values: this.values
        }
    }
    /**
     * Te da todos los datos en un formato str o String
     * @returns Collector.toJson:String
     * @example Collector.toString()
     */
    toString(){
        let i = 0
        return this.map(x => `${i+++1}.- Key: ${x.key} | Value: ${x.value}**`).join('\n')
    }
}
module.exports = Collector